/*
*  LibRAL
*  file: vcard.c
*
*  the vcard import engine v0.5 
*  
*  Copyright (C) Nicola Fragale <nicolafragale@libero.it>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <glib.h>
#include <glib-object.h>

#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "vcard.h"

#include "../../types.h"
#include "../../abook.h"
#include "../../card.h"
#include "../../personal.h"
#include "../../company.h"
#include "../../contact.h"
#include "../../work.h"
#include "../../notes.h"
#include "../../address.h"
#include "../../net.h"
#include "../../telephone.h"
#include "../../plugin.h"
#include "../../filter.h"


#define VCARD_TOKEN_DELIM ':'
#define VCARD_PARAM_DELIM ';'
#define VCARD_VALUE_DELIM '='


typedef enum {
  VCARD_START = 0,
  VCARD_BUILD_TOKEN, 
  VCARD_BUILD_PARAM, 
  VCARD_BUILD_VALUES, 
  VCARD_BUILD_DATA, 
  VCARD_END
} ParserStatus;


typedef enum {
  N = 1, 
  FN, 
  TZ, 
  END, 
  ADR, 
  TEL, 
  GEO, 
  ORG, 
  REV, 
  UID, 
  URL, 
  KEY, 
  BDAY, 
  ROLE, 
  LOGO, 
  NOTE, 
  BEGIN,
  PHOTO, 
  LABEL,
  EMAIL, 
  TITLE, 
  AGENT, 
  PROID, 
  SOUND, 
  CLASS,
  MAILER, 
  VERSION,
  NICKNAME, 
  CATEGORIES,                  // last vcard token

  KDE_EXT_Profession,          // KDE extensions
  KDE_EXT_Department,
  KDE_EXT_ManagersName,
  KDE_EXT_AssistantsName,
  KDE_EXT_Office,
  KDE_EXT_SpousesName,
  KDE_EXT_Anniversary,
  KDE_EXT_CUSTOM,

  RUBRICA_EXT_Group,           // Rubrica extensions
  RUBRICA_EXT_CompanyName,
  RUBRICA_EXT_Street,
  RUBRICA_EXT_StreetNumber,
  RUBRICA_EXT_ZipCode,
  RUBRICA_EXT_City,
  RUBRICA_EXT_Province,
  RUBRICA_EXT_Country,
  RUBRICA_EXT_Web,
  RUBRICA_EXT_Email,
  RUBRICA_EXT_Operator,
  RUBRICA_EXT_Fax,
  RUBRICA_EXT_Green,
  RUBRICA_EXT_CustomerCare,
  RUBRICA_EXT_Notes,
  RUBRICA_EXT_Department,
  RUBRICA_EXT_SubDepartment,
  RUBRICA_EXT_ManagerName,
  RUBRICA_EXT_CollaboratorName,
  RUBRICA_EXT_CollaboratorPhone,
  RUBRICA_EXT_SpouseName,
  RUBRICA_EXT_Child,
  RUBRICA_EXT_Anniversary,
  RUBRICA_EXT_Hobbies,

  EVOLUTION_EXT_File_as,       // Evolution extensions
  EVOLUTION_EXT_Office,
  EVOLUTION_EXT_Manager,
  EVOLUTION_EXT_Assistant,
  EVOLUTION_EXT_Spouse,
  EVOLUTION_EXT_Anniversary,

  MOZILLA_EXT_Html,

  BAD_TOKEN,                   // this isn't a VCard token
} RVCardTokenType;


typedef enum {
  TYPE = 1, 
  VALUE, 
  ENCODING,
  BAD_PARAM,                   // this isn't a VCard token
} RVCardParamType;


typedef enum {
  VCARD = 1,
  BINARY,
  BASIC, 
  PUBLIC, 
  PRIVATE, 
  CONFIDENTIAL,
  HOME,  
  WORK, 
  POSTAL, 
  PARCEL, 
  DOM, 
  INTL, 
  PREF,
  VOICE, 
  FAX, 
  MSG, 
  CELL, 
  PAGER, 
  BBS, 
  MODEM, 
  CAR, 
  ISDN, 
  VIDEO,  
  AOL, 
  INTERNET,
  APPLELINK, 
  ATTMAIL, 
  CIS, 
  EWORLD, 
  IBMMail, 
  MCIMAIL, 
  POWERSHARE, 
  PRODIGY, 
  TLX, 
  X400, 
  PGP, 
  X509, 
  TEXT,
  URI,
  JPEG,
  QUOTED_PRINTABLE,
  BAD_VALUE,                 // this isn't a vcard token
} RVCardValueType;



typedef struct _RVCardType {
  gchar *str;                /* token */
  gint tok;                  /* vcard decoded token */
} RVCardType;


static RVCardType tokens[] = { 
  {"N",   N},             {"FN",  FN},        {"TZ",  TZ},
  {"END", END},           {"ADR", ADR},       {"TEL", TEL},
  {"GEO", GEO},           {"ORG", ORG},       {"REV", REV},
  {"UID", UID},           {"URL", URL},       {"KEY", KEY},
  {"BDAY", BDAY},         {"ROLE", ROLE},     {"LOGO", LOGO}, 
  {"NOTE", NOTE},         {"BEGIN", BEGIN},   {"PHOTO", PHOTO},  
  {"LABEL", LABEL},       {"EMAIL", EMAIL},   {"TITLE", TITLE}, 
  {"AGENT", AGENT},       {"PROID", PROID},   {"SOUND", SOUND},  
  {"CLASS", CLASS},       {"MAILER", MAILER}, {"VERSION", VERSION}, 
  {"NICKNAME", NICKNAME}, {"CATEGORIES", CATEGORIES},

  /* kde extension.
  */
  {"X-KADDRESSBOOK-X-Profession",     KDE_EXT_Profession},
  {"X-KADDRESSBOOK-X-Department",     KDE_EXT_Department},
  {"X-KADDRESSBOOK-X-ManagersName",   KDE_EXT_ManagersName},
  {"X-KADDRESSBOOK-X-AssistantsName", KDE_EXT_AssistantsName},
  {"X-KADDRESSBOOK-X-Office",         KDE_EXT_Office},
  {"X-KADDRESSBOOK-X-SpousesName",    KDE_EXT_SpousesName},
  {"X-KADDRESSBOOK-X-Anniversary",    KDE_EXT_Anniversary},
  {"X-KADDRESSBOOK-X-CUSTOM",         KDE_EXT_CUSTOM},
  
  /* rubrica extension. 
     So rubrica can import a previously 
     exported rubrica's addressbook to vcard format
  */
  {"X-RUBRICA-X-Group",             RUBRICA_EXT_Group},
  {"X-RUBRICA-X-CompanyName",       RUBRICA_EXT_CompanyName},
  {"X-RUBRICA-X-Street",            RUBRICA_EXT_Street},
  {"X-RUBRICA-X-StreetNumber",      RUBRICA_EXT_StreetNumber},
  {"X-RUBRICA-X-ZipCode",           RUBRICA_EXT_ZipCode},
  {"X-RUBRICA-X-City",              RUBRICA_EXT_City},
  {"X-RUBRICA-X-Province",          RUBRICA_EXT_Province},
  {"X-RUBRICA-X-Country",           RUBRICA_EXT_Country},
  {"X-RUBRICA-X-Web",               RUBRICA_EXT_Web},
  {"X-RUBRICA-X-Email",             RUBRICA_EXT_Email},
  {"X-RUBRICA-X-Operator",          RUBRICA_EXT_Operator},
  {"X-RUBRICA-X-Fax",               RUBRICA_EXT_Fax},
  {"X-RUBRICA-X-Green",             RUBRICA_EXT_Green},
  {"X-RUBRICA-X-CustomerCare",      RUBRICA_EXT_CustomerCare},
  {"X-RUBRICA-X-Department",        RUBRICA_EXT_Department},
  {"X-RUBRICA-X-SubDepartment",     RUBRICA_EXT_SubDepartment},
  {"X-RUBRICA-X-ManagerName",       RUBRICA_EXT_ManagerName},
  {"X-RUBRICA-X-CollaboratorName",  RUBRICA_EXT_CollaboratorName},
  {"X-RUBRICA-X-CollaboratorPhone", RUBRICA_EXT_CollaboratorPhone},
  {"X-RUBRICA-X-SpouseName",        RUBRICA_EXT_SpouseName},
  {"X-RUBRICA-X-Child",             RUBRICA_EXT_Child},
  {"X-RUBRICA-X-Anniversary",       RUBRICA_EXT_Anniversary},

  /* evolution extensions
  */
  {"X-EVOLUTION-FILE-AS",        EVOLUTION_EXT_File_as},  
  {"X-EVOLUTION-OFFICE",         EVOLUTION_EXT_Office},
  {"X-EVOLUTION-MANAGER",        EVOLUTION_EXT_Manager},  
  {"X-EVOLUTION-ASSISTANT",      EVOLUTION_EXT_Assistant},
  {"X-EVOLUTION-SPOUSE",         EVOLUTION_EXT_Spouse},  
  {"X-EVOLUTION-ANNIVERSARY",    EVOLUTION_EXT_Anniversary},

  {"X-MOZILLA-HTML",             MOZILLA_EXT_Html},

  /*     TODO: add evolution extensions
  */
  {NULL,  BAD_TOKEN}
};
	
	
static RVCardType paramtokens[] = {
  {"type",     TYPE},
  {"value",    VALUE},   
  {"encoding", ENCODING},
  {NULL,       BAD_PARAM}
};		


static RVCardType valuetokens[] = {
  {"home",     HOME},     {"work",       WORK},       {"cell",    CELL},     
  {"fax",      FAX},      {"postal",     POSTAL},     {"parcel",  PARCEL},  
  {"dom",      DOM},      {"intl",       INTL},       {"pref",    PREF},    
  {"voice",    VOICE},    {"msg",        MSG},        {"pager",   PAGER},   
  {"bbs",      BBS},      {"modem",      MODEM},      {"car",     CAR},     
  {"isdn",     ISDN},     {"video",      VIDEO},      {"aol",     AOL},     
  {"internet", INTERNET}, {"applelink",  APPLELINK},  {"attmail", ATTMAIL},
  {"cis",      CIS},      {"eworld",     EWORLD},     {"ibmmail", IBMMail},  
  {"mcimail",  MCIMAIL},  {"powershare", POWERSHARE}, {"prodigy", PRODIGY}, 
  {"tlx",      TLX},      {"x400",       X400},       {"vcard",   VCARD},
  {"pgp",      PGP},      {"x509",       X509},       {"b",       BINARY},   
  {"binary",   BINARY},   {"text",       TEXT},       {"uri",     URI},
  {"basic",    BASIC},    {"jpeg",       JPEG},       {"public",  PUBLIC},
  {"private",  PRIVATE},  {"confidential", CONFIDENTIAL}, 
  {"quoted-printable", QUOTED_PRINTABLE},  
  {NULL,       BAD_VALUE}
};



static void r_vcard_init       (RVCard* obj);
static void r_vcard_class_init (RVCardClass* klass);

static void r_vcard_dispose    (RVCard* obj);
static void r_vcard_finalize   (RVCard* obj);



/* private functions
*/
static RCompanyCard*   r_vcard_company_new        (RVCard *vcard, RCard* card);
static RAddress*       r_vcard_company_address_new(RVCard *vcard);

static void            r_vcard_decode_token (RVCard* vcard, gchar *string);
static void            r_vcard_decode_value (RVCard* vcard, gchar *string);
static void            r_vcard_decode_param (RVCard* vcard, gchar *string);
static RDate*          r_vcard_decode_date  (gchar* buffer); 
static RTelephoneType  r_vcard_decode_telephone_type (RVCardValueType type);
static RAddressType    r_vcard_decode_address_type   (RVCardValueType type);

static RVCardTokenType r_vcard_get_token  (RVCard* vcard);
static RVCardParamType r_vcard_get_param  (RVCard* vcard);
static RVCardValueType r_vcard_get_value  (RVCard* vcard);
static gchar*          r_vcard_get_data   (RVCard* vcard);

static void            r_vcard_build_card (RVCard *vcard, RAbook* abook);

static gboolean        r_vcard_read_file    (RVCard* vcard, RAbook* abook,
					     gchar* filename);

static gboolean        r_vcard_open_file    (RAbook* vcard, gchar* filename);
static void            r_vcard_write_file   (RAbook* vcard, gchar* filename);



typedef struct _RParams {
  RVCardParamType param;
  RVCardValueType value;
} RParams;


struct _RVCardPrivate {
  gboolean have_company;
  gboolean have_company_address ;

  RVCardTokenType token;
  RVCardParamType param;
  RVCardValueType value;

  GString *data;

  GList* more;   // RParams list

  gboolean dispose_has_run;
};



RPersonalCard *card              = NULL;
RCompanyCard  *company           = NULL;
RContact      *contact           = NULL;
RWork         *work              = NULL;
RNotes        *notes             = NULL;
RAddress      *address           = NULL;
RTelephone    *tel               = NULL;
RNetAddress   *net               = NULL;
RAddress      *company_address   = NULL;
RTelephone    *company_telephone = NULL;
RNetAddress   *company_net       = NULL;

gchar** categories = NULL;
gchar** decoded;


GType
r_vcard_get_type()
{
  static GType r_vcard_type = 0;
  
  if (!r_vcard_type)
    {
      static const GTypeInfo r_vcard_info =
	{
	  sizeof(RVCardClass),
	  NULL,
	  NULL,
	  (GClassInitFunc) r_vcard_class_init,
	  NULL,
	  NULL,
	  sizeof(RVCard),
	  0,
	  (GInstanceInitFunc) r_vcard_init
	};

      r_vcard_type = g_type_register_static (G_TYPE_OBJECT,
					     "RVCard", 
					     &r_vcard_info, 0);
    }
  
  return r_vcard_type;
}


static void
r_vcard_init(RVCard* vcard)
{
  g_return_if_fail(R_VCARD(vcard) != NULL);

  vcard->private = g_new(RVCardPrivate, 1);

  vcard->fp    = NULL;

  vcard->private->have_company         = FALSE;
  vcard->private->have_company_address = FALSE;

  vcard->private->token = BAD_TOKEN;
  vcard->private->param = BAD_PARAM;
  vcard->private->value = BAD_VALUE;
  vcard->private->data  = g_string_new(NULL);
  vcard->private->more  = NULL;
  
  vcard->private->dispose_has_run = FALSE;
}


static void
r_vcard_class_init(RVCardClass* klass)
{
  GObjectClass *class;
  
  class        = G_OBJECT_CLASS (klass);

  class->dispose  = (GObjectFinalizeFunc) r_vcard_dispose;
  class->finalize = (GObjectFinalizeFunc) r_vcard_finalize;
}


static void 
r_vcard_dispose (RVCard* vcard)
{
  g_return_if_fail(IS_R_VCARD(vcard));

  if (vcard->private->dispose_has_run)
    return;

  vcard->private->dispose_has_run = TRUE;
}


static void 
r_vcard_finalize (RVCard* vcard)
{
  g_return_if_fail(IS_R_VCARD(vcard));
  
  g_free(vcard->private);
  vcard->private = NULL;
}



static RDate*
r_vcard_decode_date (gchar* buffer)
{
  g_return_val_if_fail(buffer!= NULL, NULL);

  /* date: 2001-10-21T15:53:29
           1973-03-18
  */
  RDate* date;
  struct tm tm;
  gchar *tmp;
  gboolean time2RData = TRUE;
  
  date = r_date_new();

  tmp = buffer;
  for (; *tmp; tmp++)
    if (*tmp == '-')
      time2RData = FALSE;
  
  if (!time2RData)
    {
      gchar *str, *tok;

      str = g_strdup(buffer);
      tok = strtok(str, "-");
      tm.tm_year = atoi(tok) - 1900;
      tok = strtok(NULL, "-");
      tm.tm_mon = atoi(tok) - 1;
      tok = strtok(NULL, "T");
      tm.tm_mday = atoi(tok);
      g_free(str);
      
      /*
	if (has_hour)
	{
	tok = strtok(NULL, ":");
	tm.tm_hour = atoi(tok);
	tok = strtok(NULL, ":");
	tm.tm_min = atoi(tok);
	tok = strtok(NULL, "");
	tm.tm_sec = atoi(tok);
	}
      */
    }
  else
    {
      time_t t;
      
      t = atoi(buffer);
      localtime_r (&t, &tm); 
    }
  
  r_date_know_date(date, TRUE);
  r_date_set_date(date, tm.tm_mday, tm.tm_mon, tm.tm_year);
  
  return date;
}


static void
r_vcard_decode_token(RVCard* vcard, gchar *string)
{
  gint i=0;
  
  g_return_if_fail(IS_R_VCARD(vcard));
  
  if (!string)
    vcard->private->token = BAD_TOKEN;
  
  while (tokens[i].str)
    {
      if (g_ascii_strcasecmp(string, tokens[i].str) == 0)
	{
	  vcard->private->token = tokens[i].tok;
	  break;
	}
      else
	i++;
    }
}


static void
r_vcard_decode_param(RVCard* vcard, gchar *string)
{
  gint i=0;
  
  g_return_if_fail(IS_R_VCARD(vcard));
    
  if (!string)
    vcard->private->param = BAD_PARAM;

  while (paramtokens[i].str)
    {
      if (g_ascii_strcasecmp(string, paramtokens[i].str) == 0)
	{
	  vcard->private->param = paramtokens[i].tok;
	  break;
	}
      else
	i++;
    }
}


static void 
r_vcard_decode_value(RVCard* vcard, gchar *string)
{
  gint i=0;
  
  g_return_if_fail(IS_R_VCARD(vcard));
  
  if (!string)
    vcard->private->value = BAD_VALUE;

  while (valuetokens[i].str)
    {
      if (g_ascii_strcasecmp(string, valuetokens[i].str) == 0)
	{
	  vcard->private->value = valuetokens[i].tok;
	  break;
	}
      else
	i++;
    }
}


static RCompanyCard* 
r_vcard_company_new(RVCard *vcard, RCard* card)
{
  RRef* ref;
  RCompanyCard* company = NULL;
  gulong id;
  gchar* info;
      
  vcard->private->have_company = TRUE;

  g_object_get(R_CARD(card), "card-id", &id, "card-name", &info, NULL);

  ref = r_ref_new(id);
  g_object_set(ref, "ref-info", info, NULL);

  company = r_company_card_new();
  r_card_add_ref(R_CARD(company), ref);	  
  
  return company;
}


static RAddress*
r_vcard_company_address_new(RVCard *vcard)
{
  g_return_val_if_fail(IS_R_VCARD(vcard), NULL);
  
  vcard->private->have_company_address = TRUE;

  return r_address_new();
}


static RVCardTokenType
r_vcard_get_token(RVCard* vcard)
{
  g_return_val_if_fail(IS_R_VCARD(vcard), BAD_TOKEN);

  return vcard->private->token;
}

static RVCardParamType
r_vcard_get_param(RVCard* vcard)
{
  g_return_val_if_fail(IS_R_VCARD(vcard), BAD_PARAM);

  return vcard->private->param;
}


static RVCardValueType
r_vcard_get_value(RVCard* vcard)
{
  g_return_val_if_fail(IS_R_VCARD(vcard), BAD_VALUE);

  return vcard->private->value;
}


static gchar* 
r_vcard_get_data(RVCard* vcard)
{
  g_return_val_if_fail(IS_R_VCARD(vcard), "");

  return vcard->private->data->str;
}


static RAddressType   
r_vcard_decode_address_type (RVCardValueType type)
{
  RAddressType ret = R_ADDRESS_UNKNOWN;

  switch (type)
	{
	case HOME:
	  ret = R_ADDRESS_HOME;
	  break;
	  
	case WORK:
	  ret = R_ADDRESS_WORK;
	  break;
	  
	case BAD_VALUE:
	  ret = R_ADDRESS_INVALID;
	  break;
	  
	default:
	  break;	      
	} 

  return ret;
}



static RTelephoneType 
r_vcard_decode_telephone_type(RVCardValueType type)
{
  RTelephoneType ret;
  
  switch (type)
    {	  
    case HOME:
      ret = R_TELEPHONE_HOME;
      break;
      
    case WORK:
      ret = R_TELEPHONE_WORK;
      break;
      
    case FAX:
      ret = R_TELEPHONE_FAX;
      break;
      
    case CELL:
      ret = R_TELEPHONE_CELLPHONE;
      break;
      
    case BAD_VALUE:
      ret = R_TELEPHONE_UNKNOWN;
      break;
      
      /* other telephone type */
    case PREF:
    case VOICE:
    case MSG:
    case PAGER:
    case BBS:
    case MODEM:
    case CAR:
    case ISDN:
    case VIDEO:
    default:
      ret = R_TELEPHONE_OTHER;
      break;
    }  
  
  return ret;
}


gchar*
r_vcard_validate_data(gchar* str, int n)
{
  gchar* tmp, *ret;
  
  tmp = str;
  for (; *tmp; tmp++)
    if (tmp && *tmp == ';')
      n--;      
  
  if (n > 0)
    {
      gchar* buf;
      
      buf = g_strnfill(n-1, ';');
      ret = g_strdup_printf("%s%s", str, buf);
      g_free(buf);
    }
  else
    ret = g_strdup(str);
  
  return ret;
}

void 
r_vcard_build_card(RVCard *vcard, RAbook* abook)
{
  GList* lst = NULL;
  RAddressType addtype;
  RTelephoneType type;
  RDate* date;

  gint i=0, day, month, year;

  g_return_if_fail(IS_R_VCARD(vcard));

  switch(vcard->private->token)
    {
    case BEGIN:
      card    = r_personal_card_new();
      contact = r_contact_new();
      work    = r_work_new();
      notes   = r_notes_new();
      break;
      
    case END:
      r_personal_card_set_contact(card, contact);
      r_personal_card_set_notes(card, notes);
      r_personal_card_set_work(card, work);     

      r_abook_add_card(R_ABOOK(abook), R_CARD(card));
      g_signal_emit_by_name(R_ABOOK(abook), "card_read", card, G_TYPE_POINTER);

      if (vcard->private->have_company)
	{
	  r_card_add_address(R_CARD(company), company_address);
	  vcard->private->have_company = FALSE;
	  vcard->private->have_company_address = FALSE;

	  r_abook_add_card(R_ABOOK(abook), R_CARD(company));
	  g_signal_emit_by_name(R_ABOOK(abook), "card_read", company,
				G_TYPE_POINTER);
	}
      break;

    case FN:
      /* card name */
      g_object_set(R_CARD(card), "card-name", vcard->private->data->str, NULL);
      break;
      
    case N:
      /* ord ==> surname (last name), (given) name, (middle) second name,
	         prefix, suffix (title) 
	         cognome, nome, secondo nome, prefisso, suffisso (titolo) 
      */
      if (vcard->private->data->str &&
	  (g_ascii_strcasecmp(vcard->private->data->str, "") != 0))
	{
	  gchar *data;
	  data = r_vcard_validate_data(vcard->private->data->str, 5);
	  
	  decoded = g_strsplit(data, ";", 5);   
	  g_object_set(contact, 
		       "last-name",   decoded[0],
		       "first-name",  decoded[1], 
		       "middle-name", decoded[2], 
		       "prefix",      decoded[3], 
		       "title",       decoded[4], NULL);
	  g_strfreev(decoded);
	  g_free(data);
	}
      break;

    case TZ:
      break;

    case REV:
      break;

    case ADR:
      /* mail address + mail type
      */
      address = r_address_new();

      /* pbox ==> decoded[0], extadd ==> decoded[1] */
      if (vcard->private->data->str &&
	  (g_ascii_strcasecmp(vcard->private->data->str, "") != 0))
	{
	  gchar *data;
	  
	  data = r_vcard_validate_data(vcard->private->data->str, 7);

	  decoded = g_strsplit(data, ";", 7);
	  g_object_set(address, 
		       "street",   decoded[2], 
		       "city",     decoded[3],
		       "state",    decoded[4],
		       "province", decoded[4], 
		       "zip",      decoded[5], 
		       "country",  decoded[6], NULL);
	  g_strfreev(decoded);
	  
	  addtype = r_vcard_decode_address_type(r_vcard_get_value(vcard));
	  g_object_set(address, "address-type", addtype, NULL);
	  
	  r_card_add_address(R_CARD(card), address);
	  
	  for (lst = g_list_first(vcard->private->more); lst;
	       lst = g_list_next(vcard->private->more))
	    {	  
	      RVCardValueType value;
	      
	      if (lst->data)
		value = (RVCardValueType) (((RParams*)lst->data)->value);
	      else 
		value = BAD_VALUE;
	      
	      addtype = r_vcard_decode_address_type(value);
	      
	      g_object_set(address, "address-type", addtype, NULL);
	      r_card_add_address(R_CARD(card), address);
	    }
	}
      break;
      
    case TEL:
      /* telefono + tipo */   
      tel  = r_telephone_new();      
      type = r_vcard_decode_telephone_type(r_vcard_get_value(vcard));
      g_object_set(tel, 
		   "telephone-number", r_vcard_get_data(vcard), 
		   "telephone-type",   type, 
		   NULL);
      
      r_card_add_telephone(R_CARD(card), tel);
      
      for (lst = g_list_first(vcard->private->more); lst; 
	   lst = g_list_next(vcard->private->more))
	{
	  RVCardValueType value;

	  if (lst->data)
	    value = (RVCardValueType) (((RParams*)lst->data)->value);
	  else
	    value = BAD_VALUE;
	  
	  type = r_vcard_decode_telephone_type(value);
	  
	  tel = r_telephone_new();
	  g_object_set(tel, 
		       "telephone-number", r_vcard_get_data(vcard), 
		       "telephone-type", type, 
		       NULL);
	  
	  r_card_add_telephone(R_CARD(card), tel);	  
	}
      break;
      
    case EMAIL:  
      net = r_net_address_new();
      
      g_object_set(net, 
		   "url",      r_vcard_get_data(vcard), 
		   "url-type", R_NET_ADDRESS_EMAIL, 
		   NULL);
      r_card_add_net_address(R_CARD(card), net);
      break;

    case URL:      
      net = r_net_address_new();
      
      g_object_set(net, 
		   "url",      r_vcard_get_data(vcard), 
		   "url-type", R_NET_ADDRESS_WEB,
		   NULL);
      r_card_add_net_address(R_CARD(card), net);
     break;

    case GEO:
      break;

    case BDAY:
      date = r_vcard_decode_date(r_vcard_get_data(vcard)); 
      g_object_get(date, "day",&day, "month", &month, "year", &year, NULL);
      r_contact_set_birthday(contact, day, month, year);
      break;
      
    case TITLE:
    case KDE_EXT_Profession:
      g_object_set(contact, "profession", r_vcard_get_data(vcard), NULL);
      break;

    case ORG:
      g_object_set(work, "organization", r_vcard_get_data(vcard), NULL);
      break;

    case CATEGORIES:
    case RUBRICA_EXT_Group:
      categories = g_strsplit(r_vcard_get_data(vcard), ",", 0);
      for (i = 0; categories[i]; i++)
	{
	  RGroup* group;

	  group = r_group_new();
	  g_object_set(group, 
		       "group-name", categories[i], 
		       "group-owner", "user",
		       NULL);
	  
	  r_card_add_group(R_CARD(card), group);
	}
      g_strfreev(categories); 
      break;
      
    case NOTE:
      g_object_set(notes, "other-notes", r_vcard_get_data(vcard), NULL);
      break;
      
    case KEY: 
      g_object_set(notes, "pubkey", r_vcard_get_data(vcard), NULL);
      break;

    case ROLE:
      g_object_set(work, "assignment", r_vcard_get_data(vcard), NULL);
      break;

      /*   TODO -- recuperare campi condivisi
      */
    case UID:
      break;

    case LOGO:
      break;

    case PHOTO:
      break;

    case LABEL:
      break;

    case AGENT:
      break;

    case PROID:
      break;

    case SOUND:
      break;

    case CLASS:
      break;

    case MAILER:
      break;

    case VERSION:
      break;

    case NICKNAME:
      g_object_set(contact, "nick-name", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_Department:
    case KDE_EXT_Department:
      g_object_set(work, "department", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_SubDepartment:
      g_object_set(work, "sub-department", r_vcard_get_data(vcard), NULL);
      break;

    case RUBRICA_EXT_CollaboratorName:
    case KDE_EXT_AssistantsName:
    case EVOLUTION_EXT_Assistant:
      g_object_set(work, "collaborator", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_CollaboratorPhone:
      g_object_set(work, "collaborator-phone", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_SpouseName:
    case KDE_EXT_SpousesName:
    case EVOLUTION_EXT_Spouse:
      g_object_set(notes, "has-partner", TRUE, 
		   "partner-name", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_Child:
      g_object_set(notes, "children", r_vcard_get_data(vcard), NULL);
      break;

    case RUBRICA_EXT_Anniversary:
    case KDE_EXT_Anniversary:
    case EVOLUTION_EXT_Anniversary:
      date = r_vcard_decode_date(r_vcard_get_data(vcard));
      r_notes_set_know_anniversary(notes, TRUE);
      g_object_get(date, "day",&day, "month", &month, "year", &year, NULL);
      r_notes_set_anniversary(notes, day, month, year);
    break;
      
    case RUBRICA_EXT_ManagerName:
    case KDE_EXT_ManagersName:
    case EVOLUTION_EXT_Manager:
      g_object_set(work, "manager-name", r_vcard_get_data(vcard), NULL);
      break;

    case KDE_EXT_Office:
    case EVOLUTION_EXT_Office:
      break;

    case KDE_EXT_CUSTOM:  
      break;
      
      /* Company tokens 
       */
    case RUBRICA_EXT_CompanyName:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
      
      g_object_set(company, "company-name", r_vcard_get_data(vcard), NULL); 
      g_object_set(R_CARD(company), "card-name", r_vcard_get_data(vcard),
		   NULL);
      break;

    case RUBRICA_EXT_Street:
      if (!vcard->private->have_company)
	company = r_vcard_company_new(vcard, R_CARD(card));
       
      if (!vcard->private->have_company_address)
	company_address = r_vcard_company_address_new(vcard);
 
      g_object_set(company_address, "street", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_StreetNumber:
      if (!vcard->private->have_company)
	company = r_vcard_company_new(vcard, R_CARD(card));
       
      if (!vcard->private->have_company_address)
	company_address = r_vcard_company_address_new(vcard);
 
      g_object_set(company_address, "street-number", 
		   r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_ZipCode:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
       
      if (!vcard->private->have_company_address)
	company_address = r_vcard_company_address_new(vcard);
 
      g_object_set(company_address, "zip", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_City:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
 
      if (!vcard->private->have_company_address)
	company_address = r_vcard_company_address_new(vcard);
      
      g_object_set(company_address, "city", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_Province:
      if (!vcard->private->have_company)
	company = r_vcard_company_new(vcard, R_CARD(card));
 
      if (!vcard->private->have_company_address)
	company_address = r_vcard_company_address_new(vcard);
 
      g_object_set(company_address, "province", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_Country:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
      
      if (!vcard->private->have_company_address)
	company_address = r_vcard_company_address_new(vcard);
 
      g_object_set(company_address, "country", r_vcard_get_data(vcard), NULL);
      break;
      
    case RUBRICA_EXT_Web:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
      
      company_net = r_net_address_new();
      g_object_set(company_net, 
		   "url", r_vcard_get_data(vcard), 
		   "url-type", R_NET_ADDRESS_WEB, NULL);
      r_card_add_net_address(R_CARD(company), company_net);
      break;
      
    case RUBRICA_EXT_Email:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
      
      company_net = r_net_address_new();
      g_object_set(company_net, "url", r_vcard_get_data(vcard), 
		   "url-type", R_NET_ADDRESS_EMAIL, NULL);
      r_card_add_net_address(R_CARD(company), company_net);
      break;
      
    case RUBRICA_EXT_Operator:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
	  
      company_telephone = r_telephone_new();
      g_object_set(company_telephone, 
		   "telephone-number", r_vcard_get_data(vcard), 
		   "telephone-type", R_TELEPHONE_OPERATOR, NULL);
      r_card_add_telephone(R_CARD(company), company_telephone);
      break;

    case RUBRICA_EXT_Fax:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
 
      company_telephone = r_telephone_new();
      g_object_set(company_telephone, 
		   "telephone-number", r_vcard_get_data(vcard), 
		   "telephone-type", R_TELEPHONE_FAX, NULL);
      r_card_add_telephone(R_CARD(company), company_telephone);
      break;
      
    case RUBRICA_EXT_Green:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
 
      company_telephone = r_telephone_new();
      g_object_set(company_telephone, 
		   "telephone-number", r_vcard_get_data(vcard), 
		   "telephone-type", R_TELEPHONE_GREEN, NULL);      
      r_card_add_telephone(R_CARD(company), company_telephone);   
      break;
      
    case RUBRICA_EXT_CustomerCare:
      if (!vcard->private->have_company) 
	company = r_vcard_company_new(vcard, R_CARD(card));
      
      company_telephone = r_telephone_new();
      g_object_set(company_telephone, 
		   "telephone-number", r_vcard_get_data(vcard), 
		   "telephone-type", R_TELEPHONE_CUSTOMER_CARE, NULL);
      r_card_add_telephone(R_CARD(company), company_telephone);
      break;

    case EVOLUTION_EXT_File_as:
      break;

    case BAD_TOKEN:   
    default:   
      break;
    }
}


/* Public functions
 */

RVCard* 
r_vcard_new(void)
{
  RVCard *vcard;

  vcard = g_object_new(r_vcard_get_type(), NULL);

  return (RVCard*) vcard;
}


void 
r_vcard_free (RVCard *vcard)
{
  g_return_if_fail(IS_R_VCARD(vcard));

  g_object_unref(vcard);
}



static gboolean 
r_vcard_open_file (RAbook* abook, gchar* filename)
{
  RVCard* vcard;
  
  g_return_val_if_fail(IS_R_ABOOK(abook), FALSE);
  
  if (!filename)
    {
      g_signal_emit_by_name(R_ABOOK(abook), "open_fail", 
			    NO_FILENAME, G_TYPE_INT);

      return FALSE;
    }

  if (!g_file_test(filename, G_FILE_TEST_EXISTS))
    {
      g_signal_emit_by_name(R_ABOOK(abook), "open_fail", 
			    FILE_NOT_EXIST, G_TYPE_INT);
      
      return FALSE;
    }
  
  vcard = (RVCard*) r_abook_get_plugin(abook);
  
  if (!r_vcard_read_file(vcard, abook, filename))
    {
      g_error("\nReading Vcard file");
    }
  else
    g_signal_emit_by_name(R_ABOOK(abook), "addressbook_read", 
			  NULL, G_TYPE_NONE);
  
  g_object_set(R_ABOOK(abook),
	       "addressbook-name", g_path_get_basename(filename),
	       "addressbook-path", g_path_get_dirname(filename),
	       NULL);

  return TRUE;
}

static void
write_card (RCard* card, FILE* fp)
{
  gpointer data;
  gchar *id, *name, *cardtype, *tmp, *type;
  gboolean deletable, deleted;
  gint created, changed;
  RRank rank;
  gchar *day, *month, *year;
  gchar *first, *middle, *last, *nick, *prefix, *prof, *title, *photo;

  fprintf(fp, "\n");

  fprintf(fp, "VCARD:BEGIN\n");
  fprintf(fp, "VERSION:3.0\n");

  g_object_get(card, 
	       "card-id",      &id,      "card-type",      &cardtype,
	       "card-name",    &name,    "card-deletable", &deletable, 
	       "card-deleted", &deleted, "card-rank",      &rank, 
	       "card-created", &created, "card-changed",   &changed,
	       NULL);
  
  if (name)
    fprintf(fp, "FN:%s\n", name);

  /*      Card's data (rubrica's extensions)
  */
  tmp = g_strdup_printf("%d", rank);
  fprintf(fp, "X-RUBRICA-X-Card;RANK=%s:%s;%s\n", tmp,
	  deletable ? "true" : "false",
	  deleted   ? "true" : "false");
  g_free(tmp);
  
  /*      Groups (rubrica's extensions)
  */
  data = r_card_get_group(R_CARD(card));
  for (; data; data = r_card_get_next_group(R_CARD(card)))
    {
      gchar* owner;
      gchar* pixmap = NULL;

      g_object_get(R_GROUP(data), 
		   "group-name",   &name, 
		   "group-owner",  &owner, 
		   "group-pixmap", &pixmap, NULL);
      
      fprintf(fp, "X-RUBRICA-X-Group;OWNER=%s:%s;%s\n", 
	      owner, 
	      name, 
	      pixmap);
    }


  /*      References (rubrica's extensions)
  */
  data = r_card_get_ref(R_CARD(card));
  for (; data; data = r_card_get_next_ref(R_CARD(card)))
    {
      glong id;
      gchar *info, *tmp;

      g_object_get(R_REF(data), "ref-id", &id, "ref-info", &info, NULL);
      tmp = g_strdup_printf("%ld", id);
      
      fprintf(fp, "X-RUBRICA-X-Ref;ID=%s:%s\n", tmp, info);
      g_free(tmp);
    }  

  if (g_ascii_strcasecmp(cardtype, "personal") == 0)
    {
      /*      Personal data
       */
      data = r_personal_card_get_contact(R_PERSONAL_CARD(card));
      g_object_get(R_CONTACT(data), "first-name", &first, 
		   "middle-name", &middle, "last-name", &last, 
		   "nick-name", &nick, "profession", &prof, 
		   "prefix", &prefix, "title", &title, "photo", &photo, NULL);

      if (first || middle || last || prefix || title)
	fprintf(fp, "N:%s;%s;%s;%s;%s\n", 
		last   ? last   : "", 
		first  ? first  : "",
		middle ? middle : "",
		prefix ? prefix : "",
		title  ? title  : "");
    
      day   = r_contact_get_birth_day   (R_CONTACT(data));
      month = r_contact_get_birth_month (R_CONTACT(data));
      year  = r_contact_get_birth_year  (R_CONTACT(data)); 
      if ((day   && (g_ascii_strcasecmp(day, "BadDay")   != 0)) && 
	  (month && (g_ascii_strcasecmp(day, "BadMonth") != 0)) &&
	  (year  && (g_ascii_strcasecmp(day, "BadYear")  != 0)))	
	fprintf(fp, "BDAY:%s-%s-%s\n", day, month, year);      
    }
  else
    {
      /*    Company
	    
      fprintf(fp, ":%s\n", );
      
      
      
      */
    } 
  
  /*      Addresses
  */
  data = r_card_get_address(R_CARD(card));
  for (; data; data = r_card_get_next_address(R_CARD(card)))
    {
      if (IS_R_ADDRESS(data))
	{
	  RAddressType adtype = R_ADDRESS_UNKNOWN;
	  gchar *street, *number, *city, *zip;
	  gchar *province, *state, *country;

	  g_object_get(R_ADDRESS(data), 
		       "address-type", &adtype, "street", &street, 
		       "street-number", &number,  "city", &city, 
		       "zip", &zip, "province", &province, 
		       "state", &state, "country", &country, NULL);
	  
	  type = r_address_decode_type(adtype);
	  if (street || city || state || zip || country)
	    fprintf(fp, "ADR;TYPE=%s:%s;%s;%s;%s;%s;%s;%s\n", type, "", "",
		    street  ? street  : "", 
		    city    ? city    : "", 
		    state   ? state   : "", 
		    zip     ? zip     : "", 
		    country ? country : "");
	}
    }
  
  
  /*      Net (emails, web urls)
  */
  data = r_card_get_net_address(R_CARD(card));
  for (; data; data = r_card_get_next_net_address(R_CARD(card)))
    if (IS_R_NET_ADDRESS(data))
      {
	RNetAddressType nettype;
	gchar *url;
	
	g_object_get(R_NET_ADDRESS(data), "url", &url, 
		     "url-type", &nettype, NULL);
	type = r_net_address_decode_type(nettype);
	
	switch (nettype)
	  {
	  case R_NET_ADDRESS_EMAIL: 
	    fprintf(fp, "EMAIL;INTERNET:%s\n", url);
	    break;

	  case R_NET_ADDRESS_WEB:
	    fprintf(fp, "URL:%s\n", url);
	    break;

	  case R_NET_ADDRESS_EKIGA:
	  case R_NET_ADDRESS_IRC:
	  case R_NET_ADDRESS_IRC_AIM:
	  case R_NET_ADDRESS_IRC_ICQ:
	  case R_NET_ADDRESS_IRC_JABBER:
	  case R_NET_ADDRESS_IRC_YAHOO:
	  case R_NET_ADDRESS_IRC_MSN:
	  case R_NET_ADDRESS_WORK_WEB:
	  case R_NET_ADDRESS_WORK_EMAIL:
	  case R_NET_ADDRESS_UNKNOWN:
	    fprintf(fp, "X-RUBRICA-X-URL;TYPE=%s:%s\n", type, url);
	    break;
	    
	  default:
	    break;	      
	  }
      }
     
  
  /*     Telephone
  */
  data = r_card_get_telephone(R_CARD(card));
  for (; data; data = r_card_get_next_telephone(R_CARD(card)))
    if (IS_R_TELEPHONE(data))
      {
	gchar *number, *type; 
	RTelephoneType ttype;
	
	g_object_get(R_TELEPHONE(data), "telephone-number", &number, 
		     "telephone-type", &ttype, NULL);
	type = r_telephone_decode_type(ttype);
	
	fprintf(fp, "TEL;TYPE=%s:%s\n", type, number);
      }


  /*      Work
  */
  if (g_ascii_strcasecmp(cardtype, "personal") == 0)
    {
      data = r_personal_card_get_work(R_PERSONAL_CARD(card));
      if (IS_R_WORK(data))
	{     
	  gchar *assignment, *org, *dep, *subdep;
	  gchar *manager, *mphone, *collaborator, *cphone;
	  
	  g_object_get(R_WORK(data), 
		       "assignment",     &assignment, 
		       "organization",   &org, 
		       "department",     &dep, 
		       "sub-department", &subdep, 
		       "manager-name",   &manager, 
		       "manager-phone",  &mphone, 
		       "collaborator",   &collaborator, 
		       "collaborator-phone", &cphone, NULL);
	  if (assignment)
	    fprintf(fp, "ROLE:%s\n", assignment);
	  if (org)
	    fprintf(fp, "ORG:%s\n", org);
	  if (dep)
	    fprintf(fp, "X-RUBRICA-Department:%s\n", dep);
	  if (subdep)
	    fprintf(fp, "X-RUBRICA-X-SubDepartment:%s\n", subdep);
	  if (manager)
	    fprintf(fp, "X-RUBRICA-X-Manager:%s\n", manager);
	  if (mphone)
	    fprintf(fp, "X-RUBRICA-X-ManagerTelephone:%s\n", mphone);
	  if (collaborator)
	    fprintf(fp, "X-RUBRICA-X-Collaborator:%s\n", collaborator);
	  if (cphone)
	    fprintf(fp, "X-RUBRICA-X-CollaboratorTelephone:%s\n", cphone);
	} 
    }
  
  
  /*      Notes
  */
  if (g_ascii_strcasecmp(type, "personal") == 0)
    {
      data = r_personal_card_get_notes(R_PERSONAL_CARD(card));
      if (IS_R_NOTES(data))
	{
	  gchar *other, *pkey;
	  
	  g_object_get(R_NOTES(data), 
		       "other-notes", &other, 
		       "pubkey", &pkey, NULL);
	  
	  if (other)
	    fprintf(fp, "NOTE:%s\n", other);
	  
	  if (pkey)
	    fprintf(fp, "KEY:%s\n", pkey);  
	}
    }

  fprintf(fp, "VCARD:END\n\n");
}

static gboolean 
r_vcard_save_file (RAbook* abook, gchar* filename)
{
  FILE* fp;

  g_return_val_if_fail(IS_R_ABOOK(abook), FALSE);
  g_return_val_if_fail(filename != NULL, FALSE);
  
  if (!(fp = fopen(filename, "w")))
    {
      g_warning("\nCan't write file: %s", filename);
      
      return FALSE;
    }
  
  r_abook_foreach_card (abook, (RFunc) write_card, fp);
  
  return TRUE;
}



gboolean 
r_vcard_read_file (RVCard* vcard, RAbook* abook, gchar* filename)
{
  ParserStatus status;
  GString *buffer;
  gunichar ch;

  g_return_val_if_fail(IS_R_VCARD(vcard), FALSE);
  g_return_val_if_fail(IS_R_ABOOK(abook), FALSE);
  
  vcard->fp = fopen(filename, "r");
  if (!vcard->fp)
    return FALSE;

  buffer = g_string_new(NULL);
  status = VCARD_START;
  do
    {
      ch = fgetc(vcard->fp);

      switch(status)
	{
	case VCARD_START:
	  // card_new 
	  if ((ch != '\n') && (ch != '\r') && (ch != ' '))
	    {
	      buffer = g_string_append_unichar(buffer, ch); 
	      status = VCARD_BUILD_TOKEN;
	    }
	  break;

	case VCARD_BUILD_TOKEN:
	  if (ch == VCARD_TOKEN_DELIM)
	    {
	      r_vcard_decode_token(vcard, buffer->str);
	      g_string_truncate(buffer, 0);
	      
	      status = VCARD_BUILD_DATA;
	      break;
	    }
	  
	  if (ch == VCARD_PARAM_DELIM)
	    {
	      r_vcard_decode_token(vcard, buffer->str);
	      g_string_truncate(buffer, 0);

	      status = VCARD_BUILD_PARAM;
	      break;
	    }
	  
	  if (ch == ' ')
	    break;
	  
	  if (g_unichar_isalpha(ch) || g_unichar_isgraph(ch))
	    buffer = g_string_append_unichar(buffer, ch); 	  
	  break;

	case VCARD_BUILD_PARAM: 
	  if (ch == VCARD_VALUE_DELIM) // =
	    {
	      r_vcard_decode_param(vcard, buffer->str);
	      g_string_truncate(buffer, 0);

	      status = VCARD_BUILD_VALUES;
	      break;
	    }
	  
	  if (ch == VCARD_PARAM_DELIM) // ;
	    {
	      r_vcard_decode_param(vcard, buffer->str);
	      g_string_truncate(buffer, 0);

	      status = VCARD_BUILD_PARAM;
	      break;
	    }
	   
	  if (ch == VCARD_TOKEN_DELIM) // :
	    {
	      r_vcard_decode_token(vcard, buffer->str);
	      g_string_truncate(buffer, 0);

	      status = VCARD_BUILD_DATA;
	      break;
	    }
	  
	  if (g_unichar_isalpha(ch))
	    {
	      buffer = g_string_append_unichar(buffer, ch);
	      status = VCARD_BUILD_PARAM;
	    }	  
	  break;

	case VCARD_BUILD_VALUES:
	  if (ch == VCARD_PARAM_DELIM) // ;
	    {
	      r_vcard_decode_value(vcard, buffer->str);
	      g_string_truncate(buffer, 0);

	      status = VCARD_BUILD_PARAM; 
	      break;
	    }

	  if (ch == VCARD_TOKEN_DELIM) // :
	    {
	      r_vcard_decode_value(vcard, buffer->str);
	      g_string_truncate(buffer, 0);

	      status = VCARD_BUILD_DATA;
	      break;
	    }
	  
	  buffer = g_string_append_unichar(buffer, ch);
	  break;

	case VCARD_BUILD_DATA:
	  if (ch == '\t')
	    break;

	  if (ch == '\r')
	    {
	      buffer = g_string_append_unichar(buffer, '\0');
	      break;
	    }

	  if ((ch == '\n') || (ch == EOF))
	    {
	      if (g_utf8_validate(buffer->str, -1, NULL))
		{
		  if ((vcard->private->param == ENCODING) && 
		      (vcard->private->value == QUOTED_PRINTABLE))
		    {
		      gchar *p = NULL;
		      
		      if ((p = g_strrstr(buffer->str, "=0D=0A=")) ||
			  (p = g_strrstr(buffer->str, "=0A=0D=")) ||
			  (p = g_strrstr(buffer->str, "=0A=0D"))  ||
			  (p = g_strrstr(buffer->str, "=0A")))
			*p = '\0';
		    }
		  g_string_append(vcard->private->data, buffer->str);
		  
		  r_vcard_build_card(vcard, abook);
		  g_string_truncate(buffer, 0);
		  g_string_truncate(vcard->private->data, 0);

		  status = VCARD_BUILD_TOKEN;
		}
	      break;
	    }

	  buffer = g_string_append_unichar(buffer, ch);
	  break;
	  
	case VCARD_END:
	  break;
	  
	default:
	  break;
	}

    } while(!feof(vcard->fp));

  fclose(vcard->fp);
  
  return TRUE;
}


static void            
r_vcard_write_file (RAbook* vcard, gchar* filename)
{
  r_vcard_save_file(vcard, filename);
}



void 
plugin_init (RPlugin* plugin)
{
  RPluginAction* action;
  RFilter* filter;

  g_return_if_fail(plugin != NULL);
 
  r_plugin_set_obj(plugin, r_vcard_new());
  g_object_set(plugin, 
	       "plugin-name", "vcard", 
	       "plugin-info", "This plugin manages the vcard file format",
	       "plugin-configurable", FALSE, NULL);
  
  filter = r_filter_new();
  g_object_set(filter, 
	       "filter-name", "VCard", 
	       "filter-mime", "text/x-vcard", NULL);
  r_filter_add_pattern(filter, "*.vcr");
  r_filter_add_pattern(filter, "*.vcard");
  
  r_plugin_add_filter (plugin, filter);
  
  /*
  filter = g_malloc0(sizeof(RFilter));
  filter->name = g_strdup();
  filter->pattern = g_strdup();
  r_plugin_add_filter(plugin, filter);


  filter = g_malloc0(sizeof(RFilter)); 
  filter->name = g_strdup("VCard (*.vcard)"); 
  filter->pattern = g_strdup("*.vcard"); 
  r_plugin_add_filter(plugin, filter); 

  filter = g_malloc0(sizeof(RFilter)); 
  filter->name = g_strdup("VCard (*.vcd)"); 
  filter->pattern = g_strdup("*.vcd"); 
  r_plugin_add_filter(plugin, filter); 

  filter = g_malloc0(sizeof(RFilter)); 
  filter->name = g_strdup("VCard (*.vcr)"); 
  filter->pattern = g_strdup("*.vcr"); 
  r_plugin_add_filter(plugin, filter); 
  */


  action = g_malloc(sizeof(RPluginAction));
  action->name   = g_strdup("read");
  action->handle = (gpointer) r_vcard_open_file;
  r_plugin_add_action(plugin, action);  
    
  action = g_malloc(sizeof(RPluginAction));
  action->name   = g_strdup("write");
  action->handle = (gpointer) r_vcard_write_file;;
  r_plugin_add_action(plugin, action);  

  action = g_malloc(sizeof(RPluginAction));
  action->name   = g_strdup("overwrite");
  action->handle = (gpointer) NULL;  // TODO
  r_plugin_add_action(plugin, action);  
}


void     
plugin_fini (RPlugin* plugin)
{
  // TODO: liberare la memoria e rilasciare l'handler al plugin
}
