/*
*  RAL -- Rubrica Addressbook Library
*  file: telephone.h
*  
*  Copyright (C) Nicola Fragale <nicolafragale@libero.it>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <glib.h>
#include <glib-object.h>

#include "telephone.h"
#include "utils.h"

/*   property enumetations
*/
enum {
  PHONE_NUMBER = 1,
  PHONE_TYPE
};


typedef struct _PhoneConversion {
  gint  token;
  gchar *label;
} PhoneConversion;


static PhoneConversion table[] = {
  {R_TELEPHONE_HOME,          "home"         },
  {R_TELEPHONE_WORK,          "work"         },
  {R_TELEPHONE_FAX,           "fax"          },
  {R_TELEPHONE_CELLPHONE,     "cellphone"    },
  {R_TELEPHONE_PAGER,         "pager"        },
  {R_TELEPHONE_OTHER,         "other"        },  
  {R_TELEPHONE_OPERATOR,      "operator"     },
  {R_TELEPHONE_GREEN,         "green"        },
  {R_TELEPHONE_CUSTOMER_CARE, "customer care"},
  {R_TELEPHONE_UNKNOWN,       "unknown"      }
};


struct _RTelephonePrivate {
  gchar* number;
  gint   type;

  gboolean dispose_has_run;
};


static void r_telephone_class_init (RTelephoneClass* klass);
static void r_telephone_init       (RTelephone* obj);

static void r_telephone_dispose    (RTelephone* obj);
static void r_telephone_finalize   (RTelephone* obj);


static void r_telephone_set_property (GObject* obj, guint property_id,
				      const GValue* value, GParamSpec* spec);
static void r_telephone_get_property (GObject* obj, guint property_id,
				      GValue* value, GParamSpec* spec);

GType
r_telephone_get_type()
{
  static GType tel_num_type = 0;
  
  if (!tel_num_type)
    {
      static const GTypeInfo tel_num_info =
	{
	  sizeof(RTelephoneClass),
	  NULL,
	  NULL,
	  (GClassInitFunc) r_telephone_class_init,
	  NULL,
	  NULL,
	  sizeof(RTelephone),
	  0,
	  (GInstanceInitFunc) r_telephone_init
	};

      tel_num_type = g_type_register_static (G_TYPE_OBJECT, 
					     "RTelephone",
					     &tel_num_info, 0);
    }
  
  return tel_num_type;
}


static void
r_telephone_class_init(RTelephoneClass* klass)
{
  GObjectClass *class;
  GParamSpec* pspec;
  
  class = G_OBJECT_CLASS (klass);

  class->dispose  = (GObjectFinalizeFunc) r_telephone_dispose;
  class->finalize = (GObjectFinalizeFunc) r_telephone_finalize;

  class->set_property = r_telephone_set_property;
  class->get_property = r_telephone_get_property;

  /**
   * RTelephone:telephone-number:
   *
   * contact's telephone number
   */
  pspec = g_param_spec_string("telephone-number",
			      "telephone number",
			      "contact's telephone number",
			      NULL,
			      G_PARAM_CONSTRUCT | G_PARAM_READWRITE);
  g_object_class_install_property(class, PHONE_NUMBER, pspec);
  
  /**
   * RTelephone:telephone-type:
   *
   * telephone number's type. See #RTelephoneType for valid values
   */
  pspec = g_param_spec_int("telephone-type",
			   "telephone number's type",
			   "telephone number's type",
			   R_TELEPHONE_HOME,        // min
			   R_TELEPHONE_UNKNOWN,     // max
			   R_TELEPHONE_UNKNOWN,     // default
			   G_PARAM_CONSTRUCT | G_PARAM_READWRITE);
  g_object_class_install_property(class, PHONE_TYPE, pspec);
}


static void
r_telephone_init(RTelephone* self)
{
  self->private = g_new(RTelephonePrivate, 1);

  self->private->number = NULL;

  self->private->dispose_has_run = FALSE;
}


static void 
r_telephone_dispose (RTelephone* self)
{
  g_return_if_fail(IS_R_TELEPHONE(self));
  
  if (self->private->dispose_has_run)
    return;

  self->private->dispose_has_run = TRUE;
}


static void 
r_telephone_finalize (RTelephone* self)
{
  g_return_if_fail(IS_R_TELEPHONE(self));

  r_utils_free_string(self->private->number);

  g_free(self->private);
  self->private = NULL;
}



static void 
r_telephone_set_property (GObject* obj, guint property_id,
			  const GValue* value, GParamSpec* spec)
{
  RTelephone* self = (RTelephone*) obj;
  
  switch (property_id)
    {
    case PHONE_NUMBER:
      g_free(self->private->number);
      self->private->number = g_value_dup_string(value);
      break;

    case PHONE_TYPE:
      self->private->type = g_value_get_int(value);
      break;

    default:
      break;
    }
}


static void 
r_telephone_get_property (GObject* obj, guint property_id,
			  GValue* value, GParamSpec* spec)
{
  RTelephone* self = (RTelephone*) obj;
  
  switch (property_id)
    {
    case PHONE_NUMBER:
      g_value_set_string(value, self->private->number);
      break;

    case PHONE_TYPE:
      g_value_set_int(value, self->private->type);
      break;

    default:
      break;
    }
}


/**
 * r_telephone_new
 * 
 * create a new #RTelephone
 *
 * returns: a #RTelephone*
 */
RTelephone*
r_telephone_new(void)
{
  RTelephone* telephone;

  telephone = g_object_new(r_telephone_get_type(), NULL);

  return telephone;
}


/**
 * r_telephone_free
 * @telephone: a #RTelephone
 *
 * free the #RTelephone*
 */
void
r_telephone_free(RTelephone* telephone)
{
  g_return_if_fail(IS_R_TELEPHONE(telephone));

  g_object_unref(telephone);   
}


/**
 * r_telephone_get_number
 * @telephone: a #RTelephone
 * 
 * copy the given telephone 
 *
 * returns: a new #RTelephone* 
 */
RTelephone* 
r_telephone_copy (RTelephone *telephone)
{
  RTelephone* new;
  RTelephoneType type;
  gchar *number;

  g_return_val_if_fail(IS_R_TELEPHONE(telephone), NULL);

  new = r_telephone_new();

  g_object_get(G_OBJECT(telephone), 
	       "telephone-number", &number, 
	       "telephone-type", &type, 
	       NULL); 
  g_object_set(G_OBJECT(new), 
	       "telephone-number", number, 
	       "telephone-type", type,
	       NULL);

  return new;
}



/**
 * r_telephone_decode_type
 * @type: a #RTelephoneType
 *
 * decode the given type
 *
 * returns: a gchar*
 */
gchar* 
r_telephone_decode_type (RTelephoneType type)
{
  if ((type < R_TELEPHONE_HOME) || (type > R_TELEPHONE_UNKNOWN))
    return "unknown";  

  return table[type].label;
}

/**
 * r_telephone_decode_type
 * @type: a gchar*
 *
 * encode the given string 
 *
 * returns: a #RTelephoneType 
 */
RTelephoneType 
r_telephone_encode_type (gchar* type)
{
  g_return_val_if_fail(type != NULL, R_TELEPHONE_UNKNOWN);

  if ((g_ascii_strcasecmp(type, "home") == 0))
    return R_TELEPHONE_HOME;
  
  if ((g_ascii_strcasecmp(type, "work") == 0))
    return R_TELEPHONE_WORK;

  if ((g_ascii_strcasecmp(type, "fax") == 0))
    return R_TELEPHONE_FAX;
  
  if ((g_ascii_strcasecmp(type, "cellphone") == 0))
    return R_TELEPHONE_CELLPHONE;

  if ((g_ascii_strcasecmp(type, "pager") == 0))
    return R_TELEPHONE_PAGER;

  if ((g_ascii_strcasecmp(type, "other") == 0))
    return R_TELEPHONE_OTHER;

  if ((g_ascii_strcasecmp(type, "operator") == 0))
    return R_TELEPHONE_OPERATOR;
  
  if ((g_ascii_strcasecmp(type, "green") == 0))
    return R_TELEPHONE_GREEN;
  
  if ((g_ascii_strcasecmp(type, "customer care") == 0))
    return R_TELEPHONE_CUSTOMER_CARE;
  
  return R_TELEPHONE_UNKNOWN;
}
